/*
 *   db.c  - simple database engine.
 *   Copyright (C) 1995, by Emanuel G. Noik
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "webutil.h"
#include "dbpriv.h"

/*---------------------------------------------------------------------------*/

long int dbTotalSeconds;	/* # seconds to compute and send hits */
long int dbTotalBIn;		/* # bytes read */
long int dbTotalBOut;		/* # bytes sent (hits only) */

static DB_Field schema[255];
static DB_Field query[255];

/*---------------------------------------------------------------------------*/

static int FindField( char* name, DBField fields[] )
{
   int f;
   for( f = 0; fields[f].name; f++ )
      if( strcmp( name, fields[f].name ) == 0 )
         return f;
   return -1;
}

/*---------------------------------------------------------------------------*/

static int SetValues( char* s, DB_Field f[] )
{
   char* p = s;
   char q[S];
   int i;
   for( i = 0; f[i].type != LAST; i++ )
   {
      switch( f[i].type )
      {
         case MASK:
            sscanf( p, "%[^|]|", q );
            f[i].v.m.i = strtol( q, 0, 0 );
            p += strlen( q ) + 1;
            f[i].v.m.and = 0;	/* by default OR masks */
            break;
         case STRING:
            if( sscanf( p, "%[^|]|", f[i].v.s ) == 1 )
            {
               p += strlen( f[i].v.s ) + 1;
            }
            else
            {
               f[i].v.s[0] = 0;
               p += 1;
            }
            break;
         case DATE:
            sscanf( p, "%[^|]|", q );
            sscanf( q, "%d/%d/%d %d:%d:%d",
            &f[i].v.d.m, &f[i].v.d.d, &f[i].v.d.y,
            &f[i].v.d.h, &f[i].v.d.mi, &f[i].v.d.s );
            p += strlen( q ) + 1;
            f[i].v.d.c = C_NONE;   /* by default don't check dates */
            break;
         case NUMBER:
            sscanf( p, "%[^|]|", q );
            f[i].v.n.i = strtol( q, 0, 0 );
            p += strlen( q ) + 1;
            f[i].v.n.c = C_NONE;   /* by default don't check numbers */
            break;
         case BOOL:
            sscanf( p, "%[^|]|", q );
            sscanf( q, "%d", &f[i].v.b );
            p += strlen( q ) + 1;
            break;
         case LOOKUP:
/*
            if( sscanf( p, "%[^|]|", q ) == 1 )
            {
               char* s;
               f[i].v.l.length = 0;
               p += strlen( q ) + 1;
               s = strtok( q, " " );
               while( s != 0 )
               {
                  f[i].v.l.values[f[i].v.l.length++] = atoi( s );
                  s = strtok( 0, " " );
               }
            }
            else
            {
               f[i].v.l.length = 0;
               p += 1;
            }
*/
            sscanf( p, "%[^|]|", q );
            f[i].v.l.length = 0;
            f[i].v.l.values[0] = atoi( q );
            p += strlen( q ) + 1;
            break;
         default:
            return 0;
            break;
      }
   }
   return i;
}

/*---------------------------------------------------------------------------*/

static int DateLT( DB_Date* d1, DB_Date* d2 )
{
   if( d1->y < d2->y ) return 1;
   if( d1->y == d2->y )
   {
      if( d1->m < d2->m ) return 1;
      if( d1->m == d2->m )
      {
         if( d1->d < d2->d ) return 1;
      }
   }
   return 0;
}

/*---------------------------------------------------------------------------*/

static int DCompare( DB_Date* s, DB_Date* q, DBCompar c )
{
   switch( c )
   {
      case C_EQ:   return (! (DateLT(s,q) || DateLT(q,s)));
      case C_NE:   return (DateLT(s,q) || DateLT(q,s));
      case C_LT:   return DateLT(s,q);
      case C_LE:   return (! DateLT(q,s));
      case C_GT:   return DateLT(q,s);
      case C_GE:   return (! DateLT(s,q));
      case C_NONE:
      default:     return 1;                break;
   }
}

/*---------------------------------------------------------------------------*/

static int Compare( long int s, long int q, DBCompar c )
{
   switch( c )
   {
      case C_EQ:   return (s == q) ? 1 : 0; break;
      case C_NE:   return (s != q) ? 1 : 0; break;
      case C_LT:   return (s <  q) ? 1 : 0; break;
      case C_LE:   return (s <= q) ? 1 : 0; break;
      case C_GT:   return (s >  q) ? 1 : 0; break;
      case C_GE:   return (s >= q) ? 1 : 0; break;
      case C_NONE:
      default:     return 1;                break;
   }
}

/*---------------------------------------------------------------------------*/

/*
 *  Set current record.
 */
int dbRecord( char* s )
{
   return SetValues( s, schema );
}

/*
 *  Set current query.
 */
int dbQuery( char* s )
{
   return SetValues( s, query );
}

/*---------------------------------------------------------------------------*/

/*
 *  Print current record or query.  Should do this into a buffer...
 */
int dbWrite( int record )
{
   int i;
   DB_Field* f = record ? schema : query;
   for( i = 0; f[i].type != LAST; i++ )
   {
      if( i ) printf( "," );
      switch( f[i].type )
      {
         case MASK:
            printf( "%ld", f[i].v.m.i );
            break;
         case STRING:
            if( strlen( f[i].v.s ) > 0 )
               printf( "%s", f[i].v.s );
            break;
         case DATE:
            printf( "%d/%d/%d %d:%d:%d",
            f[i].v.d.m, f[i].v.d.d, f[i].v.d.y,
            f[i].v.d.h, f[i].v.d.mi, f[i].v.d.s );
            break;
         case NUMBER:
            printf( "%ld", f[i].v.n.i );
            break;
         case BOOL:
            printf( "%d", f[i].v.b );
            break;
         case LOOKUP:
/*
            {
               int j, len = f[i].v.l.length;
               for( j = 0; j < len; j++ )
                  printf( "%d%s", f[i].v.l.values[j], j+1 == len ? "" : " " );
            }
*/
            printf( "%d", f[i].v.l.values[0] );
            break;
         default:
            return 0;
            break;
      }
   }
   printf( "\n" );
   return i;
}

/*---------------------------------------------------------------------------*/

/*
 *  Set schema from array of types.
 */
int dbSchemaT( DBField sch[] )
{
   int i;
   DBType t;
   for( i = 0; (t = sch[i].type) != LAST; i++ )
   {
      schema[i].type = t;
      query[i].type = t;
   }
   schema[i].type = LAST;
   query[i].type = LAST;
   return 1;
}

/*---------------------------------------------------------------------------*/

/*
 *  Set schema from array command line arguments.
 */
int dbSchemaC( int argc, char *argv[] )
{
   int i;
   for( i = 1; i < argc; i++ )
   {
      DBType t = (DBType)atoi( argv[i] );
      schema[i-1].type = t;
      query[i-1].type = t;
   }
   return 1;
}

/*---------------------------------------------------------------------------*/

/*
 *  Filter current record against query record
 */
int dbFilter( void )
{
   int i;
   for( i = 0; schema[i].type != LAST; i++ )
      if( dbFilterI( i ) == -1 )
         return 0;
   return 1;
}

/*
 *  Filter field 'i' of current record against query record
 * 
 *  return:	-1 no match, 0 blank, 1 match
 */
int dbFilterI( int i )
{
   long int r, q;

   switch( schema[i].type )
   {
      case MASK:
         r = schema[i].v.m.i;
         q = query[i].v.m.i;

         /* check for non-zero query mask */
         if( q == 0 ) return 0;
         if( query[i].v.m.and )
         {
            /* ANDing values -> all bits must be set */
            if( (r & q) != q ) return -1;
         }
         else
         {
            /* ORing values -> at least 1 bit must be set */
            if( (r & q) == 0 ) return -1;
         }
         break;
      case STRING:
         if( query[i].v.s[0] == 0 ) return 0;
         else
         {
            extern char *re_comp( char* );
            extern int re_exec( char* );
            char tmp[S];
            sprintf( tmp, "%s", query[i].v.s );
            lowercase( tmp );
            re_comp( tmp );
            sprintf( tmp, "%s", schema[i].v.s );
            lowercase( tmp );
            if( re_exec( tmp ) == 0 ) return -1;
         }
         break;
      case DATE:
         if( DCompare( &schema[i].v.d, &query[i].v.d, query[i].v.d.c ) == 0 )
            return -1;
         break;
      case NUMBER:
         r = schema[i].v.n.i;
         q = query[i].v.n.i;
         if( Compare( r, q, query[i].v.n.c ) == 0 )
            return -1;
         break;
      case BOOL:
         if( query[i].v.b == -1 ) return 0;
         if( query[i].v.b != schema[i].v.b ) return -1;
         break;
      case LOOKUP:
         {
            int j, len = query[i].v.l.length, matched = 0;
            if( len == 0 ) return 0;
            for( j = 0; j < len; j++ )
            {
               if( query[i].v.l.values[j] == schema[i].v.l.values[0] )
               {
                  matched = 1;
                  break;
               }
            }
            if( ! matched )
               return -1;
         }
         break;
      default:
         return 0;
         break;
   }
   return 1;
}

/*---------------------------------------------------------------------------*/

static int dbAnd( int i, int and )
{
   if( schema[i].type != MASK ) return 0;

   schema[i].v.m.and = and;
   query[i].v.m.and = and;

   return 1;
}

/*---------------------------------------------------------------------------*/

/*
 *  Set ANDing/ORing criteria of a mask field.
 */
int dbSetAndOr( char* name, char* val, DBField fields[] )
{
   int f = FindField( name, fields );
   if( f == -1 ) return 0;
   switch( val[0] )
   {
      case '1':
      case 'A': return dbAnd( f, 1 );
      case '0':
      case 'O':
      default : return dbAnd( f, 0 );
   }
}

/*---------------------------------------------------------------------------*/

static int dbCompar( int i, DBCompar c )
{
   switch( schema[i].type )
   {
      case NUMBER:
         schema[i].v.n.c = c;
         query[i].v.n.c = c;
         return 1;
      case DATE:
         schema[i].v.d.c = c;
         query[i].v.d.c = c;
         return 1;
      default:
         return 0;
   }
}

/*---------------------------------------------------------------------------*/

/*
 *  Set comparison function for numeric or date field.
 */
int dbSetCompar( char* name, char* val, DBField fields[] )
{
   char* vals[] = { "EQ", "NE", "LT", "LE", "GT", "GE", 0 };
   DBCompar c;
   int f = FindField( name, fields );
   if( f == -1 ) return 0;
   if( (c = (DBCompar)atoi( val )) == 0 ) c = stringIndex( val, vals );
   return dbCompar( f, c );
}

/*---------------------------------------------------------------------------*/

int dbUpdateQuery( char* name, char* val, DBField fields[], DBMask masks[] )
{
   int f = FindField( name, fields );

   if( f == -1 ) return 0;

   switch( fields[f].type )
   {
      case MASK:
         if( strtol( val, 0, 0 ) != 0 )
            query[f].v.m.i += strtol( val, 0, 0 );
         else
         {
            int m, found = 0;
            for( m = 0; masks[m].field; m++ )
            {
               if( strcmp( name, masks[m].field ) == 0 )
               {
                  found = 1;
                  break;
               }
            }
            if( found )
            {
               int t;
               if( (t = stringIndex( val, masks[m].values )) != -1 )
                  query[f].v.m.i += pow2( t );
               else return 0;
            }
            else return 0;
         }
         break;
      case STRING:
         lowercase( val );
         sprintf( query[f].v.s, "%s", val );
         break;
      case DATE:
         sscanf( val, "%d/%d/%d", &query[f].v.d.m, &query[f].v.d.d,
            &query[f].v.d.y );
         break;
      case BOOL:
         sscanf( val, "%d", &query[f].v.b );
         break;
      case LOOKUP:
         if( val[0] == '0' || atoi( val ) != 0 )
            query[f].v.l.values[query[f].v.l.length++] = atoi( val );
         else
         {
            int m, found = 0;
            for( m = 0; masks[m].field; m++ )
            {
               if( strcmp( name, masks[m].field ) == 0 )
               {
                  found = 1;
                  break;
               }
            }
            if( found )
            {
               int t;
               if( (t = stringIndex( val, masks[m].values )) != -1 )
                  query[f].v.l.values[query[f].v.l.length++] = t;
               else return 0;
            }
            else return 0;
         }
         break;
      default:
         break;
   }
   return 1;
}

/*---------------------------------------------------------------------------*/

int dbValue( int i, char* s )
{
   DB_Field* f = &schema[i];
   s[0] = 0;
   switch( f->type )
   {
      case MASK:
         sprintf( s, "%ld", f->v.m.i );
         break;
      case STRING:
         if( strlen( f->v.s ) > 0 )
            sprintf( s, "%s", f->v.s );
         break;
      case DATE:
         sprintf( s, "%d/%d/%d", f->v.d.m, f->v.d.d, f->v.d.y );
         break;
      case NUMBER:
         sprintf( s, "%ld", f->v.n.i );
         break;
      case BOOL:
         sprintf( s, "%d", f->v.b );
         break;
      case LOOKUP:
         sprintf( s, "%d", f->v.l.values[0] );
         break;
      default:
         return 0;
         break;
   }
   return 1;
}

/*---------------------------------------------------------------------------*/

int dbDumpHits( FILE* fp, int first, int maxhits, int* next,
int (*filter)(), void (*print)() )
{
   char buf[4096];	/* EGN: be careful about this limit */
   int i, n = 0;
   struct timeval finish, start;
   struct timezone tz;

   dbTotalBIn = 0;
   dbTotalBOut = 0;  
   gettimeofday( &start, &tz );

   /* move to first record */
   for( i = 0; i < first; i++ )
      if( fgets( buf, 4090, fp ) == 0 )
         goto eof;

   for( ; n < ( maxhits ? maxhits : 999999999 ); i++ )
   {
      if( fgets( buf, 4090, fp ) == 0 )
         goto eof;

      (void)dbRecord( buf );
      dbTotalBIn += strlen( buf );
      buf[0] = 0;

      if( filter ? filter() : dbFilter() )
      {
         extern void LoadVariables();
         LoadVariables();
         print( n, buf );
         printf( buf );
         dbTotalBOut += strlen( buf );
         n++;
      }
   }
   gettimeofday( &finish, &tz );
   dbTotalSeconds = finish.tv_sec - start.tv_sec;
   *next = i;
   return n;

   eof:
      *next = 0;
      return n;
}

/*---------------------------------------------------------------------------*/

void dbNextAnchor( char *url, int first, int maxhits, entry entries[], int m )
{
   int i;
   printf( "<A HREF=\"%s?first:%d+maxhits:%d", url, first, maxhits );

   for( i = 0; i <= m; i++ )
   {
      if( strcmp( entries[i].name, "" ) && strcmp( entries[i].val, "" ) )
      {
         printf( "+%s:%s", entries[i].name, entries[i].val );
      }
   }
   printf( "\">Next %d records</A>%c", maxhits, 10 );
}

/*---------------------------------------------------------------------------*/
